/*
 * Copyright 2017 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.record;

import java.io.IOException;
import java.io.InputStream;

/**
 * Class StructReaderStream reads a Java object from an input stream containing a native struct.
 */
public class StructReaderStream {
	private InputStream is;
	private int elmsiz;
	private int noelm;
	private StructReader sr;
	/**
	 * Create instance of StructReaderStream.
	 * @param is input stream
	 * @param elmsiz fixed size structs: size struct, variable size structs: maximum size struct
	 * @param noelm number of structs in buffer
	 */
	public StructReaderStream(InputStream is, int elmsiz, int noelm) {
		this.is = is;
		this.elmsiz = elmsiz;
		this.noelm = noelm;
		this.sr = new StructReader(new byte[0]);
	}
	/**
	 * Read.
	 * @param <T> struct type
	 * @param t class of what to read
	 * @return object read
	 * @throws RecordException if impossible to convert between types in class and struct
	 * @throws IOException if problem reading from input stream
	 */
	public <T> T read(Class<T> t) throws RecordException, IOException {
		check();
		return sr.read(t);
	}
	/**
	 * Read.
	 * @param <T> struct type
	 * @param t class of what to read
     * @param lenpvd supplies length for fields where it is not given (null indicates that it is to be ignored)
	 * @return object read
	 * @throws RecordException if impossible to convert between types in class and struct
	 * @throws IOException if problem reading from input stream
	 */
	public <T> T read(Class<T> t, LengthProvider2 lenpvd) throws RecordException, IOException {
		check();
		return sr.read(t, lenpvd);
	}
	/**
	 * More records available.
	 * @return true=more, false=no more
	 * @throws IOException if problem reading from input stream
	 */
	public boolean more() throws IOException {
		check();
		return sr.more();
	}
	/**
	 * Close underlying InputStream.
	 * @throws IOException if problem closing input stream
	 */
	public void close() throws IOException {
		is.close();
	}
	private void check() throws IOException {
		if(sr.getRemaining() < elmsiz) {
			byte[] b = fill(noelm * elmsiz);
			sr.update(b);
		}
	}
    private byte[] fill(int siz) throws IOException {
        byte[] b = new byte[siz];
        int n;
        int ix = 0;
        while((n = is.read(b, ix, b.length - ix)) > 0) {
            ix += n;
        }
        if(ix < siz) {
            byte[] tmp = new byte[ix];
            System.arraycopy(b, 0, tmp, 0, ix);
            b = tmp;
        }
        return b;
    }
}
